import { createServer, Server } from 'http';
import acme, { Client } from 'acme-client';
import fs from 'fs';

let email: string;
export const domainsToBeSigned: string[] = [];

export function addSignCert(domains: string[], payload: string) {
  domainsToBeSigned.push(...domains);
  if (!email) {
    // acme://
    email = payload.slice(7);
  }
  return domainsToBeSigned[0];
}

export async function runSignCert() {
  if (!domainsToBeSigned.length) {
    return;
  }
  console.error(`Signing certificate for ${domainsToBeSigned.join(', ')}`);
  // sign a cert with acme-client

  const [certificateKey, certificateRequest] = await acme.crypto.createCsr({
    commonName: domainsToBeSigned[0],
    altNames: domainsToBeSigned.slice(1),
  });

  // store as token
  const contentMap = new Map<string, string>();
  const server = createServer(async (req, res) => {
    // read token from http://<YOUR_DOMAIN>/.well-known/acme-challenge/<TOKEN>
    const token = req.url?.split('/.well-known/acme-challenge/').pop();
    if (!token) {
      // 404
      res.writeHead(404);
      res.end('Token not Found');
      return;
    }
    const content = contentMap.get(token);
    if (!content) {
      // 404
      res.writeHead(404);
      res.end('Content not Found');
      return;
    }
    res.writeHead(200);
    res.end(content);
  }).listen(80);
  await fs.promises.mkdir('/etc/nginx/acme', { recursive: true });
  let accountKey: Buffer;
  try {
    accountKey = await fs.promises.readFile('/etc/nginx/acme/account.pem');
  } catch (e) {
    accountKey = await acme.forge.createPrivateKey();
    await fs.promises.writeFile('/etc/nginx/acme/account.pem', accountKey);
  }
  const acmeClient = new Client({
    directoryUrl: acme.directory.letsencrypt.production,
    accountKey,
  });

  try {
    const certificate = await acmeClient.auto({
      csr: certificateRequest,
      email,
      termsOfServiceAgreed: true,
      challengePriority: ['http-01'],
      challengeCreateFn: async (authz, challenge, keyAuthorization) => {
        if (challenge.type !== 'http-01') {
          return;
        }
        // store token
        console.error(
          `Storing token for ${challenge.token}: ${keyAuthorization}`,
        );
        contentMap.set(challenge.token, keyAuthorization);
      },
      challengeRemoveFn: async (authz, challenge, keyAuthorization) => {
        // does nothing
      },
    });
    console.error('Certificate signed');
    // save certificate as fullchain.pem and key as privkey.pem
    await fs.promises.mkdir(`/etc/nginx/certs/${domainsToBeSigned[0]}`, {
      recursive: true,
    });
    await fs.promises.writeFile(
      `/etc/nginx/certs/${domainsToBeSigned[0]}/fullchain.pem`,
      certificate,
    );
    await fs.promises.writeFile(
      `/etc/nginx/certs/${domainsToBeSigned[0]}/privkey.pem`,
      certificateKey,
    );
  } catch (e) {
    console.error(
      `Failed to sign certificate for ${domainsToBeSigned.join(', ')}: ${
        e.stack
      }`,
    );
  }
  server.close();
}
